# -*- coding: utf-8 -*-

"""Amavis urls."""

from django.urls import path

from . import views

app_name = 'modoboa_amavis'

urlpatterns = [
    path('', views.index, name="index"),
    path('listing/', views._listing, name="_mail_list"),
    path('listing/page/', views.listing_page, name="mail_page"),
    path('getmailcontent/<str:mail_id>/', views.getmailcontent,
         name="mailcontent_get"),
    path('process/', views.process, name="mail_process"),
    path('delete/<str:mail_id>/', views.delete, name="mail_delete"),
    path('release/<str:mail_id>/', views.release,
         name="mail_release"),
    path('markspam/<str:mail_id>/', views.mark_as_spam,
         name="mail_mark_as_spam"),
    path('markham/<str:mail_id>/', views.mark_as_ham,
         name="mail_mark_as_ham"),
    path('learning_recipient/', views.learning_recipient,
         name="learning_recipient_set"),
    path('<str:mail_id>/', views.viewmail, name="mail_detail"),
    path('<str:mail_id>/headers/', views.viewheaders,
         name="headers_detail"),
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function, unicode_literals

import time

from django.core.management.base import BaseCommand
from django.db.models import Count

from modoboa.parameters import tools as param_tools
from ...models import Maddr, Msgrcpt, Msgs
from ...modo_extension import Amavis


class Command(BaseCommand):
    args = ""
    help = "Amavis quarantine cleanup"  # NOQA:A003

    def add_arguments(self, parser):
        """Add extra arguments to command line."""
        parser.add_argument(
            "--debug", action="store_true", default=False,
            help="Activate debug output")
        parser.add_argument(
            "--verbose", action="store_true", default=False,
            help="Display informational messages")

    def __vprint(self, msg):
        if not self.verbose:
            return
        print(msg)

    def handle(self, *args, **options):
        Amavis().load()
        if options["debug"]:
            import logging
            log = logging.getLogger("django.db.backends")
            log.setLevel(logging.DEBUG)
            log.addHandler(logging.StreamHandler())
        self.verbose = options["verbose"]

        conf = dict(param_tools.get_global_parameters("modoboa_amavis"))

        flags = ["D"]
        if conf["released_msgs_cleanup"]:
            flags += ["R"]

        self.__vprint("Deleting marked messages...")
        ids = Msgrcpt.objects.filter(rs__in=flags).values("mail_id").distinct()
        for msg in Msgs.objects.filter(mail_id__in=ids):
            if not msg.msgrcpt_set.exclude(rs__in=flags).count():
                msg.delete()

        self.__vprint(
            "Deleting messages older than {} days...".format(
                conf["max_messages_age"]))
        limit = int(time.time()) - (conf["max_messages_age"] * 24 * 3600)
        while True:
            qset = Msgs.objects.filter(
                pk__in=list(Msgs.objects.filter(time_num__lt=limit).values_list("pk", flat=True)[:5000])
            )
            if not qset.exists():
                break
            qset.delete()

        self.__vprint("Deleting unreferenced e-mail addresses...")
        while True:
            res = Maddr.objects.annotate(
                msgs_count=Count("msgs"), msgrcpt_count=Count("msgrcpt")
            ).filter(msgs_count=0, msgrcpt_count=0).values_list("id", flat=True)[:100000]
            if not res.exists():
                break
            Maddr.objects.filter(id__in=list(res)).delete()

        self.__vprint("Done.")



# -*- coding: utf-8 -*-

"""Amavis frontend default settings."""

from __future__ import unicode_literals


def apply(settings):
    """Modify settings."""
    if "DATABASE_ROUTERS" not in settings:
        settings["DATABASE_ROUTERS"] = []
    settings["DATABASE_ROUTERS"] += ["modoboa_amavis.dbrouter.AmavisRouter"]

    if "SILENCED_SYSTEM_CHECKS" not in settings:
        settings["SILENCED_SYSTEM_CHECKS"] = []
    settings["SILENCED_SYSTEM_CHECKS"] += ["fields.W342"]

    settings["AMAVIS_DEFAULT_DATABASE_ENCODING"] = "LATIN1"
    # settings["SA_LOOKUP_PATH"] = ("/usr/bin", )



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.conf import settings
from django.core.checks import Warning, register
from django.db import connections
from django.utils.translation import gettext as _

W001 = Warning(
    _("AMAVIS_DEFAULT_DATABASE_ENCODING does not match the character "
      "encoding used by the Amavis database."),
    hint=_("Check your database character encoding and set/update "
           "AMAVIS_DEFAULT_DATABASE_ENCODING."),
    id="modoboa-amavis.W001",
)

W002 = Warning(
    _("Modoboa Amavis has not been tested using the selected database engine."),
    hint=_("Try using PostgreSQL, MySQL or MariaDB."),
    id="modoboa-amavis.W002",
)


@register(deploy=True)
def check_amavis_database_encoding(app_configs, **kwargs):
    """Ensure AMAVIS_DEFAULT_DATABASE_ENCODING is set to the correct value."""
    errors = []
    db_engine = settings.DATABASES["amavis"]["ENGINE"]
    sql_query = None
    if "postgresql" in db_engine:
        sql_query = "SELECT pg_encoding_to_char(encoding) "\
            "FROM pg_database "\
            "WHERE datname = %s;"
    elif "mysql" in db_engine:
        sql_query = "SELECT DEFAULT_CHARACTER_SET_NAME "\
            "FROM INFORMATION_SCHEMA.SCHEMATA "\
            "WHERE SCHEMA_NAME = %s;"
    elif "sqlite" in db_engine:
        sql_query = "PRAGMA encoding;"
    else:
        errors.append(W002)
        return errors

    with connections["amavis"].cursor() as cursor:
        if "sqlite" in db_engine:
            cursor.execute(sql_query)
        else:
            cursor.execute(sql_query, [settings.DATABASES["amavis"]["NAME"]])
        encoding = cursor.fetchone()[0]

    if encoding.lower() != settings.AMAVIS_DEFAULT_DATABASE_ENCODING.lower():
        errors.append(W001)

    return errors



# -*- coding: utf-8 -*-

"""Management commands tests."""

from dateutil.relativedelta import relativedelta

from django.core.management import call_command
from django.utils import timezone

from modoboa.lib.tests import ModoTestCase
from .. import factories, models


class ManagementCommandTestCase(ModoTestCase):
    """Management commands tests."""

    databases = '__all__'

    def test_qcleanup(self):
        """Test qcleanup command."""
        factories.create_spam("user@test.com", rs="D")
        call_command("qcleanup")
        self.assertEqual(models.Quarantine.objects.count(), 0)
        self.assertEqual(models.Msgs.objects.count(), 0)
        self.assertEqual(models.Maddr.objects.count(), 0)
        self.assertEqual(models.Msgrcpt.objects.count(), 0)

        factories.create_spam("user@test.com", rs="D")
        msgrcpt = factories.create_spam("user@test.com", rs="R")
        call_command("qcleanup")
        # Should not raise anything
        msgrcpt.refresh_from_db()

        self.set_global_parameter("released_msgs_cleanup", True)
        call_command("qcleanup")
        with self.assertRaises(models.Msgrcpt.DoesNotExist):
            msgrcpt.refresh_from_db()

        msgrcpt = factories.create_spam("user@test.com")
        msgrcpt.mail.time_num = int(
            (timezone.now() - relativedelta(days=40)).strftime("%s"))
        msgrcpt.mail.save(update_fields=["time_num"])
        self.set_global_parameter("max_messages_age", 30)
        call_command("qcleanup")
        with self.assertRaises(models.Msgrcpt.DoesNotExist):
            msgrcpt.refresh_from_db()



# -*- coding: utf-8 -*-

"""Modoboa amavis constants."""

from __future__ import unicode_literals

import collections

from django.utils.translation import gettext_lazy as _

MESSAGE_TYPES = collections.OrderedDict((
    ("C", _("Clean")),
    ("S", _("Spam")),
    ("Y", _("Spammy")),
    ("V", _("Virus")),
    ("H", _("Bad Header")),
    ("M", _("Bad MIME")),
    ("B", _("Banned")),
    ("O", _("Over sized")),
    ("T", _("MTA error")),
    ("U", _("Unchecked")),
))

MESSAGE_TYPE_COLORS = {
    "C": "success",
    "S": "danger",
    "Y": "warning",
    "V": "danger",
    "H": "warning",
    "M": "warning",
    "B": "warning",
    "O": "warning",
    "T": "warning",
    "U": "default"
}



# -*- coding: utf-8 -*-

"""SQL connector module."""

import datetime

from django.db.models import Q

from modoboa.admin.models import Domain
from modoboa.lib.email_utils import decode

from .lib import cleanup_email_address, make_query_args
from .models import Maddr, Msgrcpt, Quarantine
from .utils import (
    ConvertFrom, fix_utf8_encoding, smart_bytes, smart_str
)


def reverse_domain_names(domains):
    """Return a list of reversed domain names."""
    return [".".join(reversed(domain.split("."))) for domain in domains]


class SQLconnector:
    """This class handles all database operations."""

    ORDER_TRANSLATION_TABLE = {
        "type": "content",
        "score": "bspam_level",
        "date": "mail__time_num",
        "subject": "mail__subject",
        "from": "mail__from_addr",
        "to": "rid__email"
    }

    QUARANTINE_FIELDS = [
        "content",
        "bspam_level",
        "rs",
        "rid__email",
        "mail__from_addr",
        "mail__subject",
        "mail__mail_id",
        "mail__time_num",
    ]

    def __init__(self, user=None, navparams=None):
        """Constructor."""
        self.user = user
        self.navparams = navparams
        self.messages = None

        self._messages_count = None
        self._annotations = {}

    def _exec(self, query, args):
        """Execute a raw SQL query.

        :param string query: query to execute
        :param list args: a list of arguments to replace in :kw:`query`
        """
        from django.db import connections, transaction

        with transaction.atomic():
            cursor = connections["amavis"].cursor()
            cursor.execute(query, args)

    def _apply_msgrcpt_simpleuser_filter(self, flt):
        """Apply specific filter for simple users."""
        if "str_email" not in self._annotations:
            self._annotations["str_email"] = ConvertFrom("rid__email")

        rcpts = [self.user.email]
        if hasattr(self.user, "mailbox"):
            rcpts += self.user.mailbox.alias_addresses

        query_rcpts = []
        for rcpt in rcpts:
            query_rcpts += make_query_args(rcpt, exact_extension=False,
                                           wildcard=".*")

        re = "(%s)" % "|".join(query_rcpts)
        return flt & Q(str_email__regex=re)

    def _apply_msgrcpt_filters(self, flt):
        """Apply filters based on user's role."""
        if self.user.role == "SimpleUsers":
            flt = self._apply_msgrcpt_simpleuser_filter(flt)
        elif not self.user.is_superuser:
            doms = Domain.objects.get_for_admin(
                self.user).values_list("name", flat=True)
            flt &= Q(rid__domain__in=reverse_domain_names(doms))
        return flt

    def _get_quarantine_content(self):
        """Fetch quarantine content.

        Filters: rs, rid, content
        """
        flt = (
            Q(rs__in=[" ", "V", "R", "p", "S", "H"])
            if self.navparams.get("viewrequests", "0") != "1" else Q(rs="p")
        )
        flt = self._apply_msgrcpt_filters(flt)
        pattern = self.navparams.get("pattern", "")
        if pattern:
            criteria = self.navparams.get("criteria")
            if criteria == "both":
                criteria = "from_addr,subject,to"
            search_flt = None
            for crit in criteria.split(","):
                if crit == "from_addr":
                    nfilter = Q(mail__from_addr__icontains=pattern)
                elif crit == "subject":
                    nfilter = Q(mail__subject__icontains=pattern)
                elif crit == "to":
                    if "str_email" not in self._annotations:
                        self._annotations["str_email"] = ConvertFrom(
                            "rid__email")
                    nfilter = Q(str_email__icontains=pattern)
                else:
                    continue
                search_flt = (
                    nfilter if search_flt is None else search_flt | nfilter
                )
            if search_flt:
                flt &= search_flt
        msgtype = self.navparams.get("msgtype", None)
        if msgtype is not None:
            flt &= Q(content=msgtype)

        flt &= Q(
            mail__in=Quarantine.objects.filter(chunk_ind=1).values("mail_id")
        )

        return (
            Msgrcpt.objects
            .annotate(**self._annotations)
            .select_related("mail", "rid")
            .filter(flt)
        )

    def messages_count(self):
        """Return the total number of messages living in the quarantine.

        We also store the built queryset for a later use.
        """
        if self.user is None or self.navparams is None:
            return None
        if self._messages_count is None:
            self.messages = self._get_quarantine_content()
            self.messages = self.messages.values(*self.QUARANTINE_FIELDS)

            order = self.navparams.get("order")
            if order is not None:
                sign = ""
                if order[0] == "-":
                    sign = "-"
                    order = order[1:]
                order = self.ORDER_TRANSLATION_TABLE[order]
                self.messages = self.messages.order_by(sign + order)

            self._messages_count = len(self.messages)

        return self._messages_count

    def fetch(self, start=None, stop=None):
        """Fetch a range of messages from the internal cache."""
        emails = []
        for qm in self.messages[start - 1:stop]:
            if qm["rs"] == "D":
                continue
            m = {
                "from": cleanup_email_address(
                    fix_utf8_encoding(qm["mail__from_addr"])
                ),
                "to": smart_str(qm["rid__email"]),
                "subject": fix_utf8_encoding(qm["mail__subject"]),
                "mailid": smart_str(qm["mail__mail_id"]),
                "date": datetime.datetime.fromtimestamp(qm["mail__time_num"]),
                "type": qm["content"],
                "score": qm["bspam_level"],
                "status": qm["rs"]
            }
            if qm["rs"] in ["", " "]:
                m["class"] = "unseen"
            elif qm["rs"] == "p":
                m["class"] = "pending"
            emails.append(m)
        return emails

    def get_recipient_message(self, address, mailid):
        """Retrieve a message for a given recipient.
        """
        assert isinstance(address, str), "address should be of type str"

        return Msgrcpt.objects\
            .annotate(str_email=ConvertFrom("rid__email"))\
            .get(mail=mailid.encode('ascii'), str_email=address)

    def set_msgrcpt_status(self, address, mailid: str, status):
        """Change the status (rs field) of a message recipient.

        :param string status: status
        """
        assert isinstance(address, str), "address should be of type str"
        addr = (
            Maddr.objects
            .annotate(str_email=ConvertFrom("email"))
            .get(str_email=address)
        )
        self._exec(
            "UPDATE msgrcpt SET rs=%s WHERE mail_id=%s AND rid=%s",
            [status, mailid.encode("ascii"), addr.id]
        )

    def get_domains_pending_requests(self, domains):
        """Retrieve pending release requests for a list of domains."""
        return Msgrcpt.objects.filter(
            rs="p", rid__domain__in=reverse_domain_names(domains))

    def get_pending_requests(self):
        """Return the number of requests currently pending."""
        rq = Q(rs="p")
        if not self.user.is_superuser:
            doms = Domain.objects.get_for_admin(self.user)
            if not doms.exists():
                return 0
            doms_q = Q(rid__domain__in=reverse_domain_names(
                doms.values_list("name", flat=True)))
            rq &= doms_q
        return Msgrcpt.objects.filter(rq).count()

    def get_mail_content(self, mailid):
        """Retrieve the content of a message."""
        content_bytes = smart_bytes("").join([
            smart_bytes(qmail.mail_text)
            for qmail in Quarantine.objects.filter(
                    mail=mailid)
        ])
        content = decode(
            content_bytes, "utf-8",
            append_to_error=("; mail_id=%s" % smart_str(mailid))
        )
        return content



# -*- coding: utf-8 -*-

from __future__ import print_function, unicode_literals

from django.contrib.sites import models as sites_models
from django.core import mail
from django.core.management.base import BaseCommand
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.translation import gettext as _

from modoboa.admin.models import Domain
from modoboa.core.models import User
from modoboa.parameters import tools as param_tools
from ...models import Msgrcpt
from ...modo_extension import Amavis
from ...sql_connector import SQLconnector


class Command(BaseCommand):
    help = "Amavis notification tool"  # NOQA:A003

    sender = None
    baseurl = None
    listingurl = None

    def add_arguments(self, parser):
        """Add extra arguments to command line."""
        parser.add_argument(
            "--smtp_host", type=str, default="localhost",
            help="The address of the SMTP server used to send notifications")
        parser.add_argument(
            "--smtp_port", type=int, default=25,
            help=("The listening port of the SMTP server used to send "
                  "notifications"))
        parser.add_argument("--verbose", action="store_true",
                            help="Activate verbose mode")

    def handle(self, *args, **options):
        Amavis().load()
        self.options = options
        self.notify_admins_pending_requests()

    def _build_message(self, rcpt, total, reqs):
        """Build new EmailMessage instance."""
        if self.options["verbose"]:
            print("Sending notification to %s" % rcpt)
        context = {
            "total": total,
            "requests": reqs,
            "baseurl": self.baseurl,
            "listingurl": self.listingurl
        }
        content = render_to_string(
            "modoboa_amavis/notifications/pending_requests.html",
            context
        )
        msg = mail.EmailMessage(
            _("[modoboa] Pending release requests"),
            content,
            self.sender,
            [rcpt]
        )
        return msg

    def notify_admins_pending_requests(self):
        self.sender = param_tools.get_global_parameter(
            "notifications_sender", app="modoboa_amavis")
        self.baseurl = "https://{}".format(
            sites_models.Site.objects.get_current().domain)
        self.listingurl = "{}{}?viewrequests=1".format(
            self.baseurl, reverse("modoboa_amavis:_mail_list"))

        messages = []
        # Check domain administators first.
        for da in User.objects.filter(groups__name="DomainAdmins"):
            if not hasattr(da, "mailbox"):
                continue
            rcpt = da.mailbox.full_address
            reqs = SQLconnector().get_domains_pending_requests(
                Domain.objects.get_for_admin(da).values_list("name", flat=True)
            )
            total = reqs.count()
            reqs = reqs.all()[:10]
            if reqs.count():
                messages.append(self._build_message(rcpt, total, reqs))

        # Then super administators.
        reqs = Msgrcpt.objects.filter(rs="p")
        total = reqs.count()
        if total:
            reqs = reqs.all()[:10]
            for su in User.objects.filter(is_superuser=True):
                if not hasattr(su, "mailbox"):
                    continue
                rcpt = su.mailbox.full_address
                messages.append(self._build_message(rcpt, total, reqs))

        # Finally, send emails.
        if not len(messages):
            return
        kwargs = {
            "host": self.options["smtp_host"],
            "port": self.options["smtp_port"]
        }
        with mail.get_connection(**kwargs) as connection:
            connection.send_messages(messages)



# -*- coding: utf-8 -*-

"""AppConfig for amavis."""

from __future__ import unicode_literals

from django.apps import AppConfig


class AmavisConfig(AppConfig):
    """App configuration."""

    name = "modoboa_amavis"
    verbose_name = "Modoboa amavis frontend"

    def ready(self):
        # Import these to force registration of checks and signals
        from . import checks  # NOQA:F401
        from . import handlers  # NOQA:F401



# -*- coding: utf-8 -*-

"""A collection of utility functions for working with the Amavis database."""

import chardet

from django.conf import settings
from django.db.models.expressions import Func
from django.utils.encoding import (
    smart_bytes as django_smart_bytes, smart_str as django_smart_str,
    smart_str as django_smart_str
)


"""
Byte fields for text data are EVIL.

MySQL uses `varbyte` fields which mysqlclient client maps to `str` (Py2) or
`bytes` (Py3), Djangos smart_* functions work as expected.

PostgreSQL uses `bytea` fields which psycopg2 maps to `memoryview`,
Djangos smart_* functions don't work as expected, you must call `tobytes()` on
the memoryview for them to work.

For convenience use smart_bytes and smart_str from this file in modoboa_amavis
to avoid any headaches.
"""


def smart_bytes(value, *args, **kwargs):
    if isinstance(value, memoryview):
        value = value.tobytes()
    return django_smart_bytes(value, *args, **kwargs)


def smart_str(value, *args, **kwargs):
    if isinstance(value, memoryview):
        value = value.tobytes()
    return django_smart_str(value, *args, **kwargs)


def smart_str(value, *args, **kwargs):
    if isinstance(value, memoryview):
        value = value.tobytes()
    return django_smart_str(value, *args, **kwargs)


def fix_utf8_encoding(value):
    """Fix utf-8 strings that contain utf-8 escaped characters.

    msgs.from_addr and msgs.subject potentialy contain badly escaped utf-8
    characters, this utility function fixes that and should be used anytime
    these fields are accesses.

    Didn't even know the raw_unicode_escape encoding existed :)
    https://docs.python.org/3/library/codecs.html?highlight=raw_unicode_escape#python-specific-encodings
    """
    assert isinstance(value, str), "value should be of type str"

    if len(value) == 0:
        # short circuit for empty strings
        return ""

    bytes_value = value.encode("raw_unicode_escape")
    try:
        value = bytes_value.decode("utf-8")
    except UnicodeDecodeError:
        encoding = chardet.detect(bytes_value)
        try:
            value = bytes_value.decode(encoding["encoding"], "replace")
        except (TypeError, UnicodeDecodeError):
            # ??? use the original value, we've done our best to try and
            # convert it to a clean utf-8 string.
            pass

    return value


class ConvertFrom(Func):
    """Convert a binary value to a string.
    Calls the database specific function to convert a binary value to a string
    using the encoding set in AMAVIS_DEFAULT_DATABASE_ENCODING.
    """

    """PostgreSQL implementation.
    See https://www.postgresql.org/docs/9.3/static/functions-string.html#FUNCTIONS-STRING-OTHER"""  # NOQA:E501
    function = "convert_from"
    arity = 1
    template = "%(function)s(%(expressions)s, '{}')".format(
        settings.AMAVIS_DEFAULT_DATABASE_ENCODING)

    def as_mysql(self, compiler, connection):
        """MySQL implementation.
        See https://dev.mysql.com/doc/refman/5.5/en/cast-functions.html#function_convert"""   # NOQA:E501
        return super().as_sql(
            compiler, connection,
            function="CONVERT",
            template="%(function)s(%(expressions)s USING {})".format(
                settings.AMAVIS_DEFAULT_DATABASE_ENCODING),
            arity=1,
        )

    def as_sqlite(self, compiler, connection):
        """SQLite implementation.
        SQLite has no equivilant function, just return the field."""
        return super().as_sql(
            compiler, connection,
            template="%(expressions)s",
            arity=1,
        )



# -*- coding: utf-8 -*-

"""Amavis tests."""

import os
from unittest import mock

from rq import SimpleWorker

from django.core import mail
from django.core.management import call_command
from django.test import override_settings
from django.urls import reverse

import django_rq

from modoboa.admin import factories as admin_factories
from modoboa.core import models as core_models
from modoboa.lib.tests import ModoTestCase
from .. import factories
from ..utils import smart_str


class TestDataMixin(object):
    """A mixin to provide test data."""

    @classmethod
    def setUpTestData(cls):  # NOQA:N802
        """Create some content."""
        super(TestDataMixin, cls).setUpTestData()
        cls.msgrcpt = factories.create_spam("user@test.com")


@override_settings(SA_LOOKUP_PATH=(os.path.dirname(__file__), ))
class ViewsTestCase(TestDataMixin, ModoTestCase):
    """Test views."""

    databases = '__all__'

    @classmethod
    def setUpTestData(cls):  # NOQA:N802
        """Create test data."""
        super(ViewsTestCase, cls).setUpTestData()
        admin_factories.populate_database()

    def tearDown(self):
        """Restore msgrcpt state."""
        self.msgrcpt.rs = " "
        self.msgrcpt.save(update_fields=["rs"])
        self.set_global_parameter("domain_level_learning", False)
        self.set_global_parameter("user_level_learning", False)

    def test_index(self):
        """Test index view."""
        url = reverse("modoboa_amavis:index")
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

        url = reverse("modoboa_amavis:_mail_list")
        response = self.ajax_get(url)
        self.assertIn("user@test.com", response["listing"])

        response = self.ajax_get("{}?pattern=pouet&criteria=both".format(url))
        self.assertIn("Empty quarantine", response["listing"])

        msgrcpt = factories.create_virus("user@test.com")
        response = self.ajax_get("{}?msgtype=V".format(url))
        self.assertIn(
            "<tr id=\"{}\">".format(smart_str(msgrcpt.mail.mail_id)),
            response["listing"])
        self.assertNotIn(
            "<tr id=\"{}\">".format(smart_str(self.msgrcpt.mail.mail_id)),
            response["listing"])

    def test_viewmail(self):
        """Test view_mail view."""
        mail_id = smart_str(self.msgrcpt.mail.mail_id)
        url = reverse("modoboa_amavis:mail_detail", args=[mail_id])
        url = "{}?rcpt={}".format(url, smart_str(self.msgrcpt.rid.email))
        response = self.ajax_get(url)
        self.assertIn("menu", response)
        content_url = reverse("modoboa_amavis:mailcontent_get", args=[mail_id])
        self.assertIn(content_url, response["listing"])

        response = self.client.get(content_url)
        self.assertEqual(response.status_code, 200)

        user = core_models.User.objects.get(username="user@test.com")
        self.client.force_login(user)
        response = self.ajax_get(url)
        self.msgrcpt.refresh_from_db()
        self.assertEqual(self.msgrcpt.rs, "V")

    def test_viewmail_selfservice(self):
        """Test view_mail in self-service mode."""
        self.client.logout()

        mail_id = smart_str(self.msgrcpt.mail.mail_id)
        url = reverse("modoboa_amavis:mail_detail", args=[mail_id])
        url = "{}?secret_id={}".format(
            url, smart_str(self.msgrcpt.mail.secret_id))
        response = self.client.get(url)
        self.assertEqual(response.status_code, 302)

        self.set_global_parameter("self_service", True)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 404)
        url = "{}&rcpt={}".format(url, smart_str(self.msgrcpt.rid.email))
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

        url = reverse("modoboa_amavis:mailcontent_get", args=[mail_id])
        url = "{}?secret_id={}".format(
            url, smart_str(self.msgrcpt.mail.secret_id))
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

    def test_viewheaders(self):
        """Test headers display."""
        mail_id = smart_str(self.msgrcpt.mail.mail_id)
        url = reverse("modoboa_amavis:headers_detail", args=[mail_id])
        response = self.client.get(url)
        self.assertContains(response, b"X-Spam-Flag: YES")

    def test_delete_msg(self):
        """Test delete view."""

        # Initiate session
        url = reverse("modoboa_amavis:_mail_list")
        response = self.ajax_get(url)

        mail_id = smart_str(self.msgrcpt.mail.mail_id)
        url = reverse("modoboa_amavis:mail_delete", args=[mail_id])
        data = {"rcpt": smart_str(self.msgrcpt.rid.email)}
        response = self.ajax_post(url, data)
        self.msgrcpt.refresh_from_db()
        self.assertEqual(self.msgrcpt.rs, "D")
        self.assertEqual(
            response["message"], "1 message deleted successfully")

    def test_delete_selfservice(self):
        """Test delete view in self-service mode."""
        self.client.logout()
        mail_id = smart_str(self.msgrcpt.mail.mail_id)
        url = reverse("modoboa_amavis:mail_delete", args=[mail_id])
        url = "{}?secret_id={}".format(
            url, smart_str(self.msgrcpt.mail.secret_id))
        self.set_global_parameter("self_service", True)
        self.ajax_get(url, status=400)
        url = "{}&rcpt={}".format(url, smart_str(self.msgrcpt.rid.email))
        self.ajax_get(url)
        self.msgrcpt.refresh_from_db()
        self.assertEqual(self.msgrcpt.rs, "D")

    @mock.patch("socket.socket")
    def test_release(self, mock_socket):
        """Test release view."""

        # Initiate session
        url = reverse("modoboa_amavis:_mail_list")
        response = self.ajax_get(url)

        mock_socket.return_value.recv.return_value = b"250 1234 Ok\r\n"
        mail_id = smart_str(self.msgrcpt.mail.mail_id)
        url = reverse("modoboa_amavis:mail_release", args=[mail_id])
        data = {"rcpt": smart_str(self.msgrcpt.rid.email)}
        response = self.ajax_post(url, data)
        self.msgrcpt.refresh_from_db()
        self.assertEqual(self.msgrcpt.rs, "R")
        self.assertEqual(
            response["message"], "1 message released successfully")

    @mock.patch("socket.socket")
    def test_release_selfservice(self, mock_socket):
        """Test release view."""
        mock_socket.return_value.recv.return_value = b"250 1234 Ok\r\n"
        self.client.logout()
        mail_id = smart_str(self.msgrcpt.mail.mail_id)
        base_url = reverse("modoboa_amavis:mail_release", args=[mail_id])
        self.set_global_parameter("self_service", True)
        # Missing rcpt -> fails
        url = "{}?secret_id=1234".format(base_url)
        self.ajax_get(url, status=400)
        # Wrong secret_id -> fails
        url = "{}?secret_id=1234&rcpt={}".format(
            base_url, smart_str(self.msgrcpt.rid.email))
        self.ajax_get(url, status=400)
        # Wrong rcpt -> fails
        url = "{}?secret_id=1234&rcpt=test@bad.com".format(base_url)
        self.ajax_get(url, status=400)
        # Request mode
        url = "{}?secret_id={}&rcpt={}".format(
            base_url, smart_str(self.msgrcpt.mail.secret_id),
            smart_str(self.msgrcpt.rid.email)
        )
        self.ajax_get(url)
        self.msgrcpt.refresh_from_db()
        self.assertEqual(self.msgrcpt.rs, "p")
        # Direct release mode
        self.set_global_parameter("user_can_release", True)
        self.ajax_get(url)
        self.msgrcpt.refresh_from_db()
        self.assertEqual(self.msgrcpt.rs, "R")

    def test_release_request(self):
        """Test release request mode."""
        user = core_models.User.objects.get(username="user@test.com")
        self.client.force_login(user)

        # Initiate session
        url = reverse("modoboa_amavis:_mail_list")
        response = self.ajax_get(url)

        mail_id = smart_str(self.msgrcpt.mail.mail_id)
        url = reverse("modoboa_amavis:mail_release", args=[mail_id])
        data = {"rcpt": smart_str(self.msgrcpt.rid.email)}
        response = self.ajax_post(url, data)
        self.msgrcpt.refresh_from_db()
        self.assertEqual(self.msgrcpt.rs, "p")
        self.assertEqual(response["message"], "1 request sent")

        # Send notification to admins
        call_command("amnotify")
        self.assertEqual(len(mail.outbox), 1)

    def _test_mark_message(self, action, status):
        """Mark message common code."""
        mail_id = smart_str(self.msgrcpt.mail.mail_id)
        url = reverse("modoboa_amavis:mail_mark_as_" + action, args=[mail_id])
        data = {"rcpt": smart_str(self.msgrcpt.rid.email)}
        response = self.ajax_post(url, data)
        self.assertEqual(
            response["message"], "Your request is being processed...")
        queue = django_rq.get_queue("default")
        worker = SimpleWorker([queue], connection=queue.connection)
        worker.work(burst=True)
        self.msgrcpt.refresh_from_db()
        self.assertEqual(self.msgrcpt.rs, status)

        self.msgrcpt.rs = " "
        self.msgrcpt.save(update_fields=["rs"])
        self.set_global_parameter("sa_is_local", False)
        response = self.ajax_post(url, data)
        self.assertEqual(
            response["message"], "Your request is being processed...")
        worker.work(burst=True)
        self.msgrcpt.refresh_from_db()
        self.assertEqual(self.msgrcpt.rs, status)

    def test_mark_as_ham(self):
        """Test mark_as_ham view."""
        self._test_mark_message("ham", "H")

        # Check domain level learning
        selection = "{} {}".format(
            smart_str(self.msgrcpt.rid.email),
            smart_str(self.msgrcpt.mail.mail_id))
        self.set_global_parameter("domain_level_learning", True)
        url = reverse("modoboa_amavis:learning_recipient_set")
        url = "{}?type=ham&selection={}".format(url, selection)
        response = self.client.get(url)
        self.assertContains(response, "value=\"global\"")
        self.assertContains(response, "value=\"domain\"")
        self.assertNotContains(response, "value=\"user\"")
        data = {
            "selection": selection,
            "ltype": "ham",
            "recipient": "domain"
        }
        response = self.ajax_post(url, data)

        # Check user level learning
        self.set_global_parameter("user_level_learning", True)
        response = self.client.get(url)
        self.assertContains(response, "value=\"user\"")
        data = {
            "selection": selection,
            "ltype": "ham",
            "recipient": "user"
        }
        response = self.ajax_post(url, data)

    def test_mark_as_spam(self):
        """Test mark_as_spam view."""
        self._test_mark_message("spam", "S")

    def test_manual_learning_as_user(self):
        """Test learning when connected as a simple user."""
        user = core_models.User.objects.get(username="user@test.com")
        self.client.force_login(user)
        mail_id = smart_str(self.msgrcpt.mail.mail_id)
        url = reverse("modoboa_amavis:mail_mark_as_ham", args=[mail_id])
        data = {"rcpt": smart_str(self.msgrcpt.rid.email)}
        response = self.ajax_post(url, data)
        self.assertEqual(response, {"status": "ok"})
        self.set_global_parameter("user_level_learning", True)
        self._test_mark_message("ham", "H")

    @mock.patch("socket.socket")
    def test_process_release(self, mock_socket):
        """Test process mode (bulk)."""
        # Initiate session
        url = reverse("modoboa_amavis:_mail_list")
        response = self.ajax_get(url)

        msgrcpt = factories.create_spam("user@test.com")
        url = reverse("modoboa_amavis:mail_process")
        selection = [
            "{} {}".format(
                smart_str(self.msgrcpt.rid.email),
                smart_str(self.msgrcpt.mail.mail_id)),
            "{} {}".format(
                smart_str(msgrcpt.rid.email),
                smart_str(msgrcpt.mail.mail_id)),
        ]
        mock_socket.return_value.recv.side_effect = [
            b"250 1234 Ok\r\n", b"250 1234 Ok\r\n"
        ]
        data = {
            "action": "release",
            "rcpt": smart_str(self.msgrcpt.rid.email),
            "selection": ",".join(selection)
        }
        self.set_global_parameter("am_pdp_mode", "inet")
        response = self.ajax_post(url, data)
        self.assertEqual(
            response["message"], "2 messages released successfully")

    def test_process_all(self):
        """Test process mode (bulk)."""
        # Initiate session
        url = reverse("modoboa_amavis:_mail_list")
        response = self.ajax_get(url)

        msgrcpt = factories.create_spam("user@test.com")
        url = reverse("modoboa_amavis:mail_process")
        selection = [
            "{} {}".format(
                smart_str(self.msgrcpt.rid.email),
                smart_str(self.msgrcpt.mail.mail_id)),
            "{} {}".format(
                smart_str(msgrcpt.rid.email),
                smart_str(msgrcpt.mail.mail_id)),
        ]
        data = {
            "rcpt": smart_str(self.msgrcpt.rid.email),
            "selection": ",".join(selection)
        }

        data["action"] = "mark_as_spam"
        response = self.ajax_post(url, data)
        self.assertEqual(
            response["message"], "Your request is being processed...")

        data["action"] = "mark_as_ham"
        response = self.ajax_post(url, data)
        self.assertEqual(
            response["message"], "Your request is being processed...")

        data = {
            "action": "delete",
            "rcpt": smart_str(self.msgrcpt.rid.email),
            "selection": ",".join(selection)
        }
        response = self.ajax_post(url, data)
        self.assertEqual(
            response["message"], "2 messages deleted successfully")



# -*- coding: utf-8 -*-

"""
Amavis forms.
"""

from django import forms
from django.utils.translation import gettext as _, gettext_lazy

from modoboa.lib import form_utils
from modoboa.parameters import forms as param_forms, tools as param_tools
from .models import Policy, Users


class DomainPolicyForm(forms.ModelForm):

    spam_subject_tag2_act = forms.BooleanField()

    class Meta:
        model = Policy
        fields = ("bypass_virus_checks", "bypass_spam_checks",
                  "spam_tag2_level", "spam_subject_tag2",
                  "spam_kill_level", "bypass_banned_checks")
        widgets = {
            "bypass_virus_checks": form_utils.HorizontalRadioSelect(),
            "bypass_spam_checks": form_utils.HorizontalRadioSelect(),
            "spam_tag2_level": forms.TextInput(
                attrs={"class": "form-control"}),
            "spam_kill_level": forms.TextInput(
                attrs={"class": "form-control"}),
            "spam_subject_tag2": forms.TextInput(
                attrs={"class": "form-control"}),
            "bypass_banned_checks": form_utils.HorizontalRadioSelect(),
        }

    def __init__(self, *args, **kwargs):
        if "instance" in kwargs:
            self.domain = kwargs["instance"]
            try:
                policy = Users.objects.get(
                    email="@%s" % self.domain.name).policy
                kwargs["instance"] = policy
            except (Users.DoesNotExist, Policy.DoesNotExist):
                del kwargs["instance"]
        super(DomainPolicyForm, self).__init__(*args, **kwargs)
        for field in self.fields.keys():
            self.fields[field].required = False

    def save(self, user, commit=True, **kwargs):
        policy = super(DomainPolicyForm, self).save(commit=False)
        for field in ["bypass_spam_checks", "bypass_virus_checks",
                      "bypass_banned_checks"]:
            if getattr(policy, field) == "":
                setattr(policy, field, None)

        if self.cleaned_data["spam_subject_tag2_act"]:
            policy.spam_subject_tag2 = None

        if commit:
            policy.save()
            try:
                u = Users.objects.get(fullname=policy.policy_name)
            except Users.DoesNotExist:
                u = Users.objects.get(email="@%s" % self.domain.name)
                u.policy = policy
                policy.save()
        return policy


class LearningRecipientForm(forms.Form):
    """A form to select the recipient of a learning request."""

    recipient = forms.ChoiceField(
        label=None, choices=[]
    )
    ltype = forms.ChoiceField(
        label="", choices=[("spam", "spam"), ("ham", "ham")],
        widget=forms.widgets.HiddenInput
    )
    selection = forms.CharField(
        label="", widget=forms.widgets.HiddenInput)

    def __init__(self, user, *args, **kwargs):
        """Constructor."""
        super(LearningRecipientForm, self).__init__(*args, **kwargs)
        choices = []
        if user.role == "SuperAdmins":
            choices.append(("global", _("Global database")))
        conf = dict(param_tools.get_global_parameters("modoboa_amavis"))
        if conf["domain_level_learning"]:
            choices.append(("domain", _("Domain's database")))
        if conf["user_level_learning"]:
            choices.append(("user", _("User's database")))
        self.fields["recipient"].choices = choices


class ParametersForm(param_forms.AdminParametersForm):
    """Extension settings."""

    app = "modoboa_amavis"

    amavis_settings_sep = form_utils.SeparatorField(
        label=gettext_lazy("Amavis settings"))

    localpart_is_case_sensitive = form_utils.YesNoField(
        label=gettext_lazy("Localpart is case sensitive"),
        initial=False,
        help_text=gettext_lazy("Value should match amavisd.conf variable %s"
                                % "$localpart_is_case_sensitive")
    )

    recipient_delimiter = forms.CharField(
        label=gettext_lazy("Recipient delimiter"),
        initial="",
        help_text=gettext_lazy("Value should match amavisd.conf variable %s"
                                % "$recipient_delimiter"),
        widget=forms.TextInput(attrs={"class": "form-control"}),
        required=False
    )

    qsettings_sep = form_utils.SeparatorField(
        label=gettext_lazy("Quarantine settings"))

    max_messages_age = forms.IntegerField(
        label=gettext_lazy("Maximum message age"),
        initial=14,
        help_text=gettext_lazy(
            "Quarantine messages maximum age (in days) before deletion"
        )
    )

    sep1 = form_utils.SeparatorField(label=gettext_lazy("Messages releasing"))

    released_msgs_cleanup = form_utils.YesNoField(
        label=gettext_lazy("Remove released messages"),
        initial=False,
        help_text=gettext_lazy(
            "Remove messages marked as released while cleaning up "
            "the database"
        )
    )

    am_pdp_mode = forms.ChoiceField(
        label=gettext_lazy("Amavis connection mode"),
        choices=[("inet", "inet"), ("unix", "unix")],
        initial="unix",
        help_text=gettext_lazy("Mode used to access the PDP server"),
        widget=form_utils.HorizontalRadioSelect()
    )

    am_pdp_host = forms.CharField(
        label=gettext_lazy("PDP server address"),
        initial="localhost",
        help_text=gettext_lazy("PDP server address (if inet mode)"),
        widget=forms.TextInput(attrs={"class": "form-control"})
    )

    am_pdp_port = forms.IntegerField(
        label=gettext_lazy("PDP server port"),
        initial=9998,
        help_text=gettext_lazy("PDP server port (if inet mode)")
    )

    am_pdp_socket = forms.CharField(
        label=gettext_lazy("PDP server socket"),
        initial="/var/amavis/amavisd.sock",
        help_text=gettext_lazy("Path to the PDP server socket (if unix mode)")
    )

    user_can_release = form_utils.YesNoField(
        label=gettext_lazy("Allow direct release"),
        initial=False,
        help_text=gettext_lazy(
            "Allow users to directly release their messages")
    )

    self_service = form_utils.YesNoField(
        label=gettext_lazy("Enable self-service mode"),
        initial=False,
        help_text=gettext_lazy("Activate the 'self-service' mode")
    )

    notifications_sender = forms.EmailField(
        label=gettext_lazy("Notifications sender"),
        initial="notification@modoboa.org",
        help_text=gettext_lazy(
            "The e-mail address used to send notitications")
    )

    lsep = form_utils.SeparatorField(label=gettext_lazy("Manual learning"))

    manual_learning = form_utils.YesNoField(
        label=gettext_lazy("Enable manual learning"),
        initial=True,
        help_text=gettext_lazy(
            "Allow super administrators to manually train Spamassassin"
        )
    )

    sa_is_local = form_utils.YesNoField(
        label=gettext_lazy("Is Spamassassin local?"),
        initial=True,
        help_text=gettext_lazy(
            "Tell if Spamassassin is running on the same server than modoboa"
        )
    )

    default_user = forms.CharField(
        label=gettext_lazy("Default user"),
        initial="amavis",
        help_text=gettext_lazy(
            "Name of the user owning the default bayesian database"
        )
    )

    spamd_address = forms.CharField(
        label=gettext_lazy("Spamd address"),
        initial="127.0.0.1",
        help_text=gettext_lazy("The IP address where spamd can be reached")
    )

    spamd_port = forms.IntegerField(
        label=gettext_lazy("Spamd port"),
        initial=783,
        help_text=gettext_lazy("The TCP port spamd is listening on")
    )

    domain_level_learning = form_utils.YesNoField(
        label=gettext_lazy("Enable per-domain manual learning"),
        initial=False,
        help_text=gettext_lazy(
            "Allow domain administrators to train Spamassassin "
            "(within dedicated per-domain databases)"
        )
    )

    user_level_learning = form_utils.YesNoField(
        label=gettext_lazy("Enable per-user manual learning"),
        initial=False,
        help_text=gettext_lazy(
            "Allow simple users to personally train Spamassassin "
            "(within a dedicated database)"
        )
    )

    visibility_rules = {
        "am_pdp_host": "am_pdp_mode=inet",
        "am_pdp_port": "am_pdp_mode=inet",
        "am_pdp_socket": "am_pdp_mode=unix",

        "sa_is_local": "manual_learning=True",
        "default_user": "manual_learning=True",
        "spamd_address": "sa_is_local=False",
        "spamd_port": "sa_is_local=False",
        "domain_level_learning": "manual_learning=True",
        "user_level_learning": "manual_learning=True"
    }


class UserSettings(param_forms.UserParametersForm):
    """Per-user settings."""

    app = "modoboa_amavis"

    dsep = form_utils.SeparatorField(label=gettext_lazy("Display"))

    messages_per_page = forms.IntegerField(
        initial=40,
        label=gettext_lazy("Number of displayed emails per page"),
        help_text=gettext_lazy(
            "Set the maximum number of messages displayed in a page")
    )



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.test import SimpleTestCase

from modoboa_amavis.utils import fix_utf8_encoding


class FixUTF8EncodingTests(SimpleTestCase):

    """Tests for modoboa_amavis.utils.fix_utf8_encoding()."""

    def test_4_byte_unicode(self):
        value = "\xf0\x9f\x99\x88"
        expected_output = "\U0001f648"  # == See No Evil Moneky
        output = fix_utf8_encoding(value)
        self.assertEqual(output, expected_output)

    def test_truncated_4_byte_unicode(self):
        value = "\xf0\x9f\x99"
        expected_output = "\xf0\x9f\x99"
        output = fix_utf8_encoding(value)
        self.assertEqual(output, expected_output)



# -*- coding: utf-8 -*-

"""
An email representation based on a database record.
"""

from html2text import HTML2Text

from django.template.loader import render_to_string

from modoboa.lib.email_utils import Email
from .sql_connector import SQLconnector
from .utils import fix_utf8_encoding, smart_str


class SQLemail(Email):

    """The SQL version of the Email class."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.qtype = ""
        self.qreason = ""

        qreason = self.msg["X-Amavis-Alert"]
        if qreason:
            if "," in qreason:
                self.qtype, qreason = qreason.split(",", 1)
            elif qreason.startswith("BAD HEADER SECTION "):
                # Workaround for amavis <= 2.8.0 :p
                self.qtype = "BAD HEADER SECTION"
                qreason = qreason[19:]

            qreason = " ".join([x.strip() for x in qreason.splitlines()])
            self.qreason = qreason

    def _fetch_message(self):
        return SQLconnector().get_mail_content(self.mailid)

    @property
    def body(self):
        if self._body is None:
            super().body
            self._body = fix_utf8_encoding(self._body)

        # if there's no plain text version available attempt to make one by
        # sanitising the html version. The output isn't always pretty but it
        # is readable, better than a blank screen and helps the user decide
        # if the message is spam or ham.
        if self.dformat == "plain" and not self.contents["plain"] \
                and self.contents["html"]:
            h = HTML2Text()
            h.ignore_tables = True
            h.images_to_alt = True
            mail_text = h.handle(self.contents["html"])
            self.contents["plain"] = self._post_process_plain(
                smart_str(mail_text))
            self._body = self.viewmail_plain()
            self._body = fix_utf8_encoding(self._body)

        return self._body

    def render_headers(self, **kwargs):
        context = {
            "qtype": self.qtype,
            "qreason": self.qreason,
            "headers": self.headers,
        }
        return render_to_string("modoboa_amavis/mailheaders.html", context)



# -*- coding: utf-8 -*-

"""Tests for sql_email."""

import os

from django.test import TestCase
from django.utils.encoding import smart_str

from ..sql_email import SQLemail

SAMPLES_DIR = os.path.realpath(
    os.path.join(os.path.dirname(__file__), "sample_messages"))


class EmailTestImplementation(SQLemail):

    def _fetch_message(self):
        message_path = os.path.join(SAMPLES_DIR, "%s-input.txt" % self.mailid)
        assert os.path.isfile(message_path), "%s does not exist." % message_path

        with open(message_path, "rb") as fp:
            mail_text = fp.read()

        return mail_text


class EmailTests(TestCase):
    """Tests for modoboa_amavis.sql_email.SQLEmail

    When writing new sample messages use the following naming convention for
    the sample files stored in sample_messages:

    input:  {message_id}-input.txt
    output: {message_id}-output-{dformat}_{no,}links.txt
    """

    def _get_expected_output(self, message_id, **kwargs):
        ext = kwargs["dformat"] if "dformat" in kwargs else "plain"
        ext += "_links" if "links" in kwargs and kwargs["links"] else "_nolinks"
        message_path = os.path.join(SAMPLES_DIR,
                                    "%s-output-%s.txt" % (message_id, ext))
        assert os.path.isfile(message_path), "%s does not exist." % message_path

        with open(message_path, "rb") as fp:
            # output should always be unicode (py2) or str (py3)
            mail_text = smart_str(fp.read())

        return mail_text

    def _test_email(self, message_id, **kwargs):
        """Boiler plate code for testing e-mails."""
        expected_output = self._get_expected_output(message_id, **kwargs)
        output = EmailTestImplementation(message_id, **kwargs).body
        self.assertEqual(output, expected_output)

    def test_amavis_aleart_header(self):
        email = EmailTestImplementation("quarantined")
        self.assertEqual(email.qtype, "BAD HEADER SECTION")
        self.assertEqual(email.qreason,
                         "Non-encoded non-ASCII data (and not UTF-8) (char 85 "
                         "hex): Subject: I think I saw you in my dreams\\x{85}")

    def test_email_multipart_with_no_text(self):
        """for a multipart message without a text/plain part convert the
           text/html to text/plain"""
        self._test_email("quarantined")



# -*- coding: utf-8 -*-

"""Custom test runner."""

from __future__ import unicode_literals

from django.apps import apps
from django.test.runner import DiscoverRunner


class UnManagedModelTestRunner(DiscoverRunner):
    """
    Test runner that automatically makes all unmanaged models in your Django
    project managed for the duration of the test run.
    Many thanks to the Caktus Group: http://bit.ly/1N8TcHW
    """

    unmanaged_models = []

    def setup_test_environment(self, *args, **kwargs):
        """Mark modoboa_amavis models as managed during testing
        During database setup migrations are only run for managed models"""
        for m in apps.get_models():
            if m._meta.app_label == "modoboa_amavis":
                self.unmanaged_models.append(m)
                m._meta.managed = True
        super(UnManagedModelTestRunner, self).setup_test_environment(
            *args, **kwargs)

    def teardown_test_environment(self, *args, **kwargs):
        """Revert modoboa_amavis models to unmanaged"""
        super(UnManagedModelTestRunner, self).teardown_test_environment(
            *args, **kwargs)
        for m in self.unmanaged_models:
            m._meta.managed = False



# -*- coding: utf-8 -*-

"""Amavis factories."""

from __future__ import unicode_literals

import datetime
import time

import factory

from . import models
from .utils import smart_bytes

SPAM_BODY = """X-Envelope-To: <{rcpt}>
X-Envelope-To-Blocked: <{rcpt}>
X-Quarantine-ID: <nq6ekd4wtXZg>
X-Spam-Flag: YES
X-Spam-Score: 1000.985
X-Spam-Level: ****************************************************************
X-Spam-Status: Yes, score=1000.985 tag=2 tag2=6.31 kill=6.31
    tests=[ALL_TRUSTED=-1, GTUBE=1000, PYZOR_CHECK=1.985]
    autolearn=no autolearn_force=no
Received: from demo.modoboa.org ([127.0.0.1])
    by localhost (demo.modoboa.org [127.0.0.1]) (amavisd-new, port 10024)
    with ESMTP id nq6ekd4wtXZg for <user@demo.local>;
    Thu,  9 Nov 2017 15:59:52 +0100 (CET)
Received: from demo.modoboa.org (localhost [127.0.0.1])
    by demo.modoboa.org (Postfix) with ESMTP
    for <user@demo.local>; Thu,  9 Nov 2017 15:59:52 +0100 (CET)
Content-Type: text/plain; charset="utf-8"
MIME-Version: 1.0
Content-Transfer-Encoding: base64
Subject: Sample message
From: {sender}
To: {rcpt}
Message-ID: <151023959268.5550.5713670714483771838@demo.modoboa.org>
Date: Thu, 09 Nov 2017 15:59:52 +0100

This is the GTUBE, the
        Generic
        Test for
        Unsolicited
        Bulk
        Email

If your spam filter supports it, the GTUBE provides a test by which you
can verify that the filter is installed correctly and is detecting incoming
spam. You can send yourself a test mail containing the following string of
characters (in upper case and with no white spaces and line breaks):

XJS*C4JDBQADN1.NSBN3*2IDNEN*GTUBE-STANDARD-ANTI-UBE-TEST-EMAIL*C.34X

You should send this test mail from an account outside of your network.
"""

VIRUS_BODY = """Subject: Virus Test Message (EICAR)
MIME-Version: 1.0
Content-Type: multipart/mixed; boundary="huq684BweRXVnRxX"
Content-Disposition: inline
Date: Sun, 06 Nov 2011 10:08:18 -0800


--huq684BweRXVnRxX
Content-Type: text/plain; charset=us-ascii
Content-Disposition: inline


This is a virus test message. It contains an attached file 'eicar.com',
which contains the EICAR virus <http://eicar.org/86-0-Intended-use.html>
test pattern.


--huq684BweRXVnRxX
Content-Type: application/x-msdos-program
Content-Disposition: attachment; filename="eicar.com"
Content-Transfer-Encoding: quoted-printable

X5O!P%@AP[4\PZX54(P^)7CC)7}$EICAR-STANDARD-ANTIVIRUS-TEST-FILE!$H+H*=0A
--huq684BweRXVnRxX--
"""


class MaddrFactory(factory.django.DjangoModelFactory):
    """Factory for Maddr."""

    class Meta:
        model = models.Maddr
        django_get_or_create = ("email", )

    id = factory.Sequence(lambda n: n)  # NOQA:A003
    email = factory.Sequence(lambda n: "user_{}@domain.test".format(n))
    domain = "test.domain"


class MsgsFactory(factory.django.DjangoModelFactory):
    """Factory for Mailaddr."""

    class Meta:
        model = models.Msgs

    mail_id = factory.Sequence(lambda n: "mailid{}".format(n).encode("ascii"))
    secret_id = factory.Sequence(lambda n: smart_bytes("id{}".format(n)))
    sid = factory.SubFactory(MaddrFactory)
    client_addr = "127.0.0.1"
    originating = "Y"
    dsn_sent = "N"
    subject = factory.Sequence(lambda n: "Test message {}".format(n))
    time_num = factory.LazyAttribute(lambda o: int(time.time()))
    time_iso = factory.LazyAttribute(
        lambda o: datetime.datetime.fromtimestamp(o.time_num).isoformat())
    size = 100


class MsgrcptFactory(factory.django.DjangoModelFactory):
    """Factory for Msgrcpt."""

    class Meta:
        model = models.Msgrcpt

    rseqnum = 1
    is_local = "Y"
    bl = "N"
    wl = "N"
    mail = factory.SubFactory(MsgsFactory)
    rid = factory.SubFactory(MaddrFactory)


class QuarantineFactory(factory.django.DjangoModelFactory):
    """Factory for Quarantine."""

    class Meta:
        model = models.Quarantine

    chunk_ind = 1
    mail = factory.SubFactory(MsgsFactory)


def create_quarantined_msg(rcpt, sender, rs, body, **kwargs):
    """Create a quarantined msg."""
    msgrcpt = MsgrcptFactory(
        rs=rs,
        rid__email=rcpt,
        rid__domain="com.test",  # FIXME
        mail__sid__email=smart_bytes(sender),
        mail__sid__domain="",  # FIXME
        **kwargs
    )
    QuarantineFactory(
        mail=msgrcpt.mail,
        mail_text=smart_bytes(SPAM_BODY.format(rcpt=rcpt, sender=sender))
    )
    return msgrcpt


def create_spam(rcpt, sender="spam@evil.corp", rs=" "):
    """Create a spam."""
    body = SPAM_BODY.format(rcpt=rcpt, sender=sender)
    body += "fóó bár"
    return create_quarantined_msg(
        rcpt, sender, rs, body, bspam_level=999.0, content="S")


def create_virus(rcpt, sender="virus@evil.corp", rs=" "):
    """Create a virus."""
    return create_quarantined_msg(rcpt, sender, rs, VIRUS_BODY, content="V")



# -*- coding: utf-8 -*-

from __future__ import unicode_literals


class AmavisRouter(object):

    """A router to control all database operations on models in
    the amavis application"""

    def db_for_read(self, model, **hints):
        """Point all operations on amavis models to 'amavis'."""
        if model._meta.app_label == "modoboa_amavis":
            return "amavis"
        return None

    def db_for_write(self, model, **hints):
        """Point all operations on amavis models to 'amavis'."""
        if model._meta.app_label == "modoboa_amavis":
            return "amavis"
        return None

    def allow_relation(self, obj1, obj2, **hints):
        """Allow any relation if a model in amavis is involved."""
        if obj1._meta.app_label == "modoboa_amavis" \
           or obj2._meta.app_label == "modoboa_amavis":
            return True
        return None

    def allow_migrate(self, db, app_label, model=None, **hints):
        """
        Make sure the auth app only appears in the 'amavis'
        database.
        """
        if app_label == "modoboa_amavis":
            # modoboa_amavis migrations should be created in the amavis
            # database.
            return (db == "amavis")
        elif db == "amavis":
            # Don't create non modoboa_amavis migrations in the amavis database.
            return False
        return None



"""Async tasks."""

from typing import List

from django.utils.translation import ngettext

from modoboa.core import models as core_models

from .lib import SpamassassinClient
from .sql_connector import SQLconnector


def manual_learning(user_pk: int,
                    mtype: str,
                    selection: List[str],
                    recipient_db: str):
    """Task to trigger manual learning for given selection."""
    user = core_models.User.objects.get(pk=user_pk)
    connector = SQLconnector()
    saclient = SpamassassinClient(user, recipient_db)
    for item in selection:
        rcpt, mail_id = item.split()
        content = connector.get_mail_content(mail_id.encode("ascii"))
        result = saclient.learn_spam(rcpt, content) if mtype == "spam" \
            else saclient.learn_ham(rcpt, content)
        if not result:
            break
        connector.set_msgrcpt_status(
            rcpt, mail_id, mtype[0].upper()
        )
    if saclient.error is None:
        saclient.done()
        message = ngettext("%(count)d message processed successfully",
                           "%(count)d messages processed successfully",
                           len(selection)) % {"count": len(selection)}
    else:
        message = saclient.error



# -*- coding: utf-8 -*-

"""
Django settings for test_project project.

Generated by 'django-admin startproject' using Django 1.11.8.

For more information on this file, see
https://docs.djangoproject.com/en/1.11/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.11/ref/settings/
"""

import os
from logging.handlers import SysLogHandler

from modoboa.test_settings import *  # noqa

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.realpath(os.path.dirname(os.path.dirname(__file__)))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.11/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "w537@nm@5n)=+e%-7*z-jxf21a#0k%uv^rbu**+cj4=_u57e(8"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = "DEBUG" in os.environ

TEMPLATE_DEBUG = DEBUG

ALLOWED_HOSTS = [
    "127.0.0.1",
    "localhost",
]

SITE_ID = 1

# Security settings

X_FRAME_OPTIONS = "SAMEORIGIN"

# Application definition

INSTALLED_APPS = (
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.sites",
    "django.contrib.staticfiles",
    "reversion",
    "ckeditor",
    "ckeditor_uploader",
    "rest_framework",
    "rest_framework.authtoken",
    'django_otp',
    'django_otp.plugins.otp_totp',
    'django_otp.plugins.otp_static',
)

# A dedicated place to register Modoboa applications
# Do not delete it.
# Do not change the order.
MODOBOA_APPS = (
    "modoboa",
    "modoboa.core",
    "modoboa.lib",
    "modoboa.admin",
    "modoboa.transport",
    "modoboa.relaydomains",
    "modoboa.limits",
    "modoboa.parameters",
    "modoboa.dnstools",
    "modoboa.maillog",
    "modoboa.pdfcredentials",
    "modoboa.dmarc",
    "modoboa.imap_migration",
    # Modoboa extensions here.
    "modoboa_amavis",
)

INSTALLED_APPS += MODOBOA_APPS

AUTH_USER_MODEL = "core.User"

MIDDLEWARE = (
    "x_forwarded_for.middleware.XForwardedForMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    'django_otp.middleware.OTPMiddleware',
    'modoboa.core.middleware.TwoFAMiddleware',
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.locale.LocaleMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "modoboa.core.middleware.LocalConfigMiddleware",
    "modoboa.lib.middleware.AjaxLoginRedirect",
    "modoboa.lib.middleware.CommonExceptionCatcher",
    "modoboa.lib.middleware.RequestCatcherMiddleware",
)

AUTHENTICATION_BACKENDS = (
    # 'modoboa.lib.authbackends.LDAPBackend',
    # 'modoboa.lib.authbackends.SMTPBackend',
    "django.contrib.auth.backends.ModelBackend",
)

# SMTP authentication
# AUTH_SMTP_SERVER_ADDRESS = 'localhost'
# AUTH_SMTP_SERVER_PORT = 25
# AUTH_SMTP_SECURED_MODE = None  # 'ssl' or 'starttls' are accepted


TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.template.context_processors.i18n",
                "django.template.context_processors.media",
                "django.template.context_processors.static",
                "django.template.context_processors.tz",
                "django.contrib.messages.context_processors.messages",
                "modoboa.core.context_processors.top_notifications",
            ],
            "debug": TEMPLATE_DEBUG,
        },
    },
]

ROOT_URLCONF = "test_project.urls"

WSGI_APPLICATION = "test_project.wsgi.application"


# Internationalization
# https://docs.djangoproject.com/en/1.11/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_L10N = True

USE_TZ = True

# Default primary key field type
# https://docs.djangoproject.com/en/3.2/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = 'django.db.models.BigAutoField'

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.11/howto/static-files/

STATIC_URL = "/sitestatic/"
STATIC_ROOT = os.path.join(BASE_DIR, "sitestatic")
STATICFILES_DIRS = (
    # os.path.join(BASE_DIR, '..', 'modoboa', 'bower_components'),
)

MEDIA_URL = "/media/"
MEDIA_ROOT = os.path.join(BASE_DIR, "media")

# Rest framework settings

REST_FRAMEWORK = {
    "DEFAULT_AUTHENTICATION_CLASSES": (
        "rest_framework.authentication.TokenAuthentication",
        "rest_framework.authentication.SessionAuthentication",
    ),
}

# Modoboa settings
# MODOBOA_CUSTOM_LOGO = os.path.join(MEDIA_URL, "custom_logo.png")

# DOVECOT_LOOKUP_PATH = ('/path/to/dovecot', )

MODOBOA_API_URL = "https://api.modoboa.org/1/"

# REDIS

REDIS_HOST = os.environ.get('REDIS_HOST', '127.0.0.1')
REDIS_PORT = os.environ.get('REDIS_PORT', 6379)
REDIS_QUOTA_DB = 0
REDIS_URL = 'redis://{}:{}/{}'.format(REDIS_HOST, REDIS_PORT, REDIS_QUOTA_DB)

# RQ

RQ_QUEUES = {
    'default': {
        'URL': REDIS_URL,
    },
}

# Password validation
# https://docs.djangoproject.com/en/1.11/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",  # NOQA:E501
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",  # NOQA:E501
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",  # NOQA:E501
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",  # NOQA:E501
    },
    {
        "NAME": "modoboa.core.password_validation.ComplexityValidator",
        "OPTIONS": {
            "upper": 1,
            "lower": 1,
            "digits": 1,
            "specials": 0
        }
    },
]

# CKeditor

CKEDITOR_UPLOAD_PATH = "uploads/"

CKEDITOR_IMAGE_BACKEND = "pillow"

CKEDITOR_RESTRICT_BY_USER = True

CKEDITOR_BROWSE_SHOW_DIRS = True

CKEDITOR_ALLOW_NONIMAGE_FILES = False

CKEDITOR_CONFIGS = {
    "default": {
        "allowedContent": True,
        "toolbar": "Modoboa",
        "width": None,
        "toolbar_Modoboa": [
            ["Bold", "Italic", "Underline"],
            ["JustifyLeft", "JustifyCenter", "JustifyRight", "JustifyBlock"],
            ["BidiLtr", "BidiRtl", "Language"],
            ["NumberedList", "BulletedList", "-", "Outdent", "Indent"],
            ["Undo", "Redo"],
            ["Link", "Unlink", "Anchor", "-", "Smiley"],
            ["TextColor", "BGColor", "-", "Source"],
            ["Font", "FontSize"],
            ["Image", ],
            ["SpellChecker"]
        ],
    },
}

# Logging configuration

LOGGING = {
    "version": 1,
    "formatters": {
        "syslog": {
            "format": "%(name)s: %(levelname)s %(message)s"
        },
    },
    "handlers": {
        "syslog-auth": {
            "class": "logging.handlers.SysLogHandler",
            "facility": SysLogHandler.LOG_AUTH,
            "formatter": "syslog"
        },
        "modoboa": {
            "class": "modoboa.core.loggers.SQLHandler",
        }
    },
    "loggers": {
        "modoboa.auth": {
            "handlers": ["syslog-auth", "modoboa"],
            "level": "INFO",
            "propagate": False
        },
        "modoboa.admin": {
            "handlers": ["modoboa"],
            "level": "INFO",
            "propagate": False
        }
    }
}

# Load settings from extensions
try:
    from modoboa_amavis import settings as modoboa_amavis_settings
    modoboa_amavis_settings.apply(globals())
except AttributeError:
    from modoboa_amavis.settings import *  # noqa


MIGRATION_MODULES = {
    "modoboa_amavis": None
}

TEST_RUNNER = "modoboa_amavis.test_runners.UnManagedModelTestRunner"

# We force sqlite backend for tests because the generated database is
# not the same as the one provided by amavis...
DATABASES.update({  # noqa
    "amavis": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "amavis.db",
        "PORT": "",
        "ATOMIC_REQUESTS": True,
    },
})
# sqlite defaults to UTF-8
AMAVIS_DEFAULT_DATABASE_ENCODING = "UTF-8"



# -*- coding: utf-8 -*-

"""The amavis frontend of Modoboa."""

from __future__ import unicode_literals

from pkg_resources import DistributionNotFound, get_distribution

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    pass

default_app_config = "modoboa_amavis.apps.AmavisConfig"



# -*- coding: utf-8 -*-

from django.test import TestCase
from django.test.utils import override_settings

from ..checks import settings_checks


class CheckSessionCookieSecureTest(TestCase):

    databases = '__all__'

    @override_settings(AMAVIS_DEFAULT_DATABASE_ENCODING="LATIN-1")
    def test_amavis_database_encoding_incorrect(self):
        """
        If AMAVIS_DEFAULT_DATABASE_ENCODING is incorrect provide one warning.
        """
        self.assertEqual(
            settings_checks.check_amavis_database_encoding(None),
            [settings_checks.W001]
        )

    @override_settings(AMAVIS_DEFAULT_DATABASE_ENCODING="UTF-8")
    def test_amavis_database_encoding_correct(self):
        """
        If AMAVIS_DEFAULT_DATABASE_ENCODING is correct, there's no warning.
        """
        self.assertEqual(
            settings_checks.check_amavis_database_encoding(None),
            []
        )



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.test import SimpleTestCase

from modoboa.lib.tests import ModoTestCase
from modoboa_amavis.lib import cleanup_email_address, make_query_args


class MakeQueryArgsTests(ModoTestCase):

    """Tests for modoboa_amavis.lib.make_query_args()."""

    def test_simple_email_address_001(self):
        """Check case insensitive email address without recipient delimiter."""
        self.set_global_parameter("localpart_is_case_sensitive", False)
        self.set_global_parameter("recipient_delimiter", "")
        address = "User+Foo@sub.exAMPLE.COM"
        expected_output = [
            "user+foo@sub.example.com",
        ]
        output = make_query_args(address)
        self.assertEqual(output, expected_output)

    def test_simple_email_address_002(self):
        """Check case sensitive email address with recipient delimiter."""
        self.set_global_parameter("localpart_is_case_sensitive", True)
        self.set_global_parameter("recipient_delimiter", "+")
        address = "User+Foo@sub.exAMPLE.COM"
        expected_output = [
            "User+Foo@sub.exAMPLE.COM",
            "User+Foo@sub.example.com",
            "User@sub.example.com",
        ]
        output = make_query_args(address)
        self.assertEqual(output, expected_output)

    def test_simple_email_address_003(self):
        """Check email address with recipient delimiter wildcard."""
        self.set_global_parameter("localpart_is_case_sensitive", False)
        self.set_global_parameter("recipient_delimiter", "+")
        address = "User+Foo@sub.exAMPLE.COM"
        expected_output = [
            "user+foo@sub.example.com",
            "user+.*@sub.example.com",
            "user@sub.example.com",
        ]
        output = make_query_args(address, exact_extension=False, wildcard=".*")
        self.assertEqual(output, expected_output)

    def test_simple_email_address_004(self):
        """Check domain search."""
        self.set_global_parameter("localpart_is_case_sensitive", False)
        self.set_global_parameter("recipient_delimiter", "+")
        address = "User+Foo@sub.exAMPLE.COM"
        expected_output = [
            "user+foo@sub.example.com",
            "user+.*@sub.example.com",
            "user@sub.example.com",
            "@sub.example.com",
            "@.",
        ]
        output = make_query_args(
            address, exact_extension=False, wildcard=".*", domain_search=True)
        self.assertEqual(output, expected_output)

    def test_simple_email_address_idn(self):
        """Check email address with international domain name."""
        self.set_global_parameter("localpart_is_case_sensitive", False)
        self.set_global_parameter("recipient_delimiter", "")
        address = "Pingüino@Pájaro.Niño.exAMPLE.COM"
        expected_output = [
            "Pingüino@Pájaro.Niño.exAMPLE.COM",
            "pingüino@xn--pjaro-xqa.xn--nio-8ma.example.com",
        ]
        output = make_query_args(address)
        self.assertEqual(output, expected_output)


class FixUTF8EncodingTests(SimpleTestCase):

    """Tests for modoboa_amavis.lib.cleanup_email_address()."""

    def test_value_with_newline(self):
        value = "\"John Smith\" <john.smith@example.com>\n"
        expected_output = "John Smith <john.smith@example.com>"
        output = cleanup_email_address(value)
        self.assertEqual(output, expected_output)

    def test_no_name(self):
        value = "<john.smith@example.com>"
        expected_output = "john.smith@example.com"
        output = cleanup_email_address(value)
        self.assertEqual(output, expected_output)
